"""
This file is the main API to use Interferometry without the PyQt5 UI.

TODO:
 * This should be sliced into separate files to reduce length.
 * SPP should be a different class, allow to pass in multiple datasets.
 * FFT rewrite
 * Docstrings
 * Debug
"""
import sys
import warnings

warnings.filterwarnings("ignore", message="invalid value encountered in sqrt")
warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")

import numpy as np
import matplotlib.pyplot as plt 
import pandas as pd
from scipy.fftpack import fftshift

from pysprint.api.dataset_base import DatasetBase
from pysprint.core.evaluate import (min_max_method, cff_method, fft_method,
	cut_gaussian, ifft_method, spp_method, args_comp, gaussian_window)
from pysprint.core.dataedits import savgol, find_peak, convolution, cut_data, cwt
from pysprint.core.generator import generatorFreq, generatorWave
from pysprint.core.optimizer import FitOptimizer
from pysprint.core.peak import EditPeak
from pysprint.core.normalize import DraggableEnvelope
from pysprint.utils import (print_disp, run_from_ipython, 
	findNearest as find_nearest, calc_envelope, _maybe_increase_before_cwt)


__all__ = ['Generator', 'Dataset', 'MinMaxMethod', 'CosFitMethod', 'SPPMethod', 'FFTMethod']

C_LIGHT = 299.793 # nm/fs

# setting up the IPython notebook
if run_from_ipython():
	plt.rcParams['figure.figsize'] = [15, 5]


class DatasetError(Exception):
	"""
	This error is raised when invalid type of data encountered when initializing 
	a dataset or inherited object.
	"""
	pass


class InterpolationWarning(Warning):
	"""
	This warning is raised when a function applies linear interpolation on the data.
	"""
	pass


class FourierWarning(Warning):
	"""
	This warning is raised when FFT is called first instead of IFFT.
	Later on it will be improved. 
	For more details see help(pysprint.FFTMethod.calculate)
	"""
	pass


class BaseApp(DatasetBase):
	def __init__(self):
		super.__init__()

	def run(self):
		"""
		Opens up the GUI with the loaded data.
		"""
		from pysprint.logic import MainProgram
		try:
			from PyQt5 import QtWidgets
		except ImportError:
			print('PyQt5 is essential for the UI. Use the API instead.')
		print('Building up UI..')
		app = QtWidgets.QApplication(sys.argv)
		main_app = MainProgram()
		main_app.showMaximized()
		main_app.a = self.x
		main_app.b = self.y
		main_app.samY = self.sam
		main_app.refY = self.ref
		if main_app.settings.value('show') == 'True':
			main_app.msgbox.exec_()
			if main_app.cb.isChecked():
				main_app.settings.setValue('show', False)
		else:
			pass
		main_app.redraw_graph()
		main_app.fill_table()
		main_app.track_stats()
		sys.exit(app.exec_())
		
	@property
	def data(self):
		raise NotImplementedError

	def show(self):
		raise NotImplementedError

	

class Generator(BaseApp):
	"""
	Basic dataset generator.
	"""
	def __init__(self, start, stop, center, delay=0,
		GD=0, GDD=0, TOD=0, FOD=0, QOD=0, resolution=0.1,
	 	delimiter=',', pulse_width=10, normalize=False, chirp=0):
		self.start = start
		self.stop = stop
		self.center = center
		self.delay = delay
		self.GD = GD
		self.GDD = GDD
		self.TOD = TOD
		self.FOD = FOD
		self.QOD = QOD
		self.resolution = resolution
		self.delimiter = delimiter
		self.pulse_width = pulse_width
		self.chirp = chirp
		self.normalize = normalize
		self.x = np.array([])
		self.y = np.array([])
		self.ref = np.array([])
		self.sam = np.array([])
		self.plotwidget = plt
		self.is_wave = False
		
	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		return f'''Generator({self.start}, {self.stop}, {self.center}, delay = {self.delay},
				   GD={self.GD}, GDD={self.GDD}, TOD={self.TOD}, FOD={self.FOD}, QOD={self.QOD}, resolution={self.resolution}, 
				   delimiter={self.delimiter}, pulse_width={self.pulseWidth}, normalize={self.normalize})'''

	def _check_norm(self):
		"""
		Do the normalization when we can.
		"""
		if len(self.ref) != 0:
			self._y =  (self.y - self.ref - self.sam)/(2*np.sqrt(self.sam*self.ref))

	def generate_freq(self):
		"""
		Use this to generate the spectrogram in ang. frequency domain.
		"""
		self.x, self.y, self.ref, self.sam = generatorFreq(self.start, self.stop, self.center, self.delay, self.GD,
			self.GDD, self.TOD, self.FOD, self.QOD,
			self.resolution, self.delimiter, self.pulse_width, self.normalize, self.chirp)

	def generate_wave(self):
		"""
		Use this to generate the spectrogram in wavelength domain.
		"""
		self.is_wave = True
		self.x, self.y, self.ref, self.sam = generatorWave(self.start, self.stop, self.center, self.delay, self.GD,
			self.GDD, self.TOD, self.FOD, self.QOD,
			self.resolution, self.delimiter, self.pulse_width, self.normalize, self.chirp)

	def show(self):
		"""
		Draws the plot of the generated data.
		"""
		self._check_norm()
		if np.iscomplexobj(self.y):
			self.plotwidget.plot(self.x, np.abs(self.y))
		else:   
			try:
				self.plotwidget.plot(self.x, self._y, 'r')
			except Exception:
				self.plotwidget.plot(self.x, self.y, 'r')
		self.plotwidget.grid()
		self.plotwidget.show()

	def save(self, name, path=None):
		"""
		Saves the generated dataset with numpy.savetxt.

		Parameters:
		----------

		name: string
			Name of the output file. You shouldn't include the .txt at the end.

		path: string, default is None
			You can also specify the save path.
			e.g path='C:/examplefolder'
		"""
		if path is None:
			np.savetxt('{}.txt'.format(name), np.transpose([self.x, self.y, self.ref, self.sam]), delimiter=self.delimiter)
			print(f'Successfully saved as {name}')
		else:
			np.savetxt(
				'{}/{}.txt'.format(path, name),
				 np.transpose([self.x, self.y, self.ref, self.sam]),
				 delimiter = self.delimiter
				 )
			print(f'Successfully saved as {name}')

	def _phase(self, j):
		if self.is_wave:
			lam = np.arange(self.start, self.stop + self.resolution, self.resolution) 
			omega = (2 * np.pi * C_LIGHT) / lam 
			omega0 = (2 * np.pi * C_LIGHT) / self.center 
			j = omega - omega0
		else:
			lamend = (2 * np.pi * C_LIGHT) / self.start
			lamstart = (2 * np.pi * C_LIGHT) / self.stop
			lam = np.arange(lamstart, lamend + self.resolution, self.resolution)
			omega = (2 * np.pi * C_LIGHT) / lam 
			j = omega - self.center
		return (j + self.delay * j + j * self.GD + (self.GDD / 2) * j ** 2 
			   + (self.TOD / 6)* j ** 3 + (self.FOD / 24) * j ** 4 + (self.QOD / 120) * j ** 5)

	def phase_graph(self):
		"""
		Plots the spectrogram along with the spectral phase.
		"""
		self._check_norm()
		self.fig, self.ax = self.plotwidget.subplots(2, 1, figsize=(8, 7))
		self.plotwidget.subplots_adjust(top=0.95)
		self.fig.canvas.set_window_title('Spectrum and phase')
		try:
			self.ax[0].plot(self.x, self._y, 'r')
		except Exception:
			self.ax[0].plot(self.x, self.y, 'r')
		try:
			self.ax[1].plot(self.x, self._phase(self.x))
		except Exception:
			raise ValueError('''The spectrum is not generated yet.
			Use self.generate_freq() on frequency domain or self.generate_wave() on wavelength domain.''')
		self.ax[0].set(xlabel="Frequency/Wavelength", ylabel="Intensity")
		self.ax[1].set(xlabel="Frequency/Wavelength", ylabel="$\Phi $[rad]")
		self.ax[0].grid()
		self.ax[1].grid()
		self.plotwidget.show()

	@property
	def data(self):
		"""
		Unpacks the generated data.
		If arms are given it returns x, y, reference_y, sample_y
		Else returns x, y
		"""
		if len(self.ref) == 0:
			return self.x, self.y
		return self.x, self.y, self.ref, self.sam


class Dataset(BaseApp):
	"""
	Base class for the evaluating methods.
	"""
	_metadata = None

	def __init__(self, x, y, ref=None, sam=None):
		self.x = x
		self.y = y
		if ref is None:
			self.ref = []
		else:
			self.ref = ref 
		if sam is None:
			self.sam = []
		else:
			self.sam = sam
		self._is_normalized = False
		if not isinstance(self.x, np.ndarray):
			try:
				self.x = np.array(self.x)
				self.x.astype(float)
			except ValueError:
				raise DatasetError('Invalid type of data')
		if not isinstance(self.y, np.ndarray):
			try:
				self.y = np.array(self.y)
				self.y.astype(float)
			except ValueError:
				raise DatasetError('Invalid type of data')
		if not isinstance(self.ref, np.ndarray):
			try:
				self.ref = np.array(self.ref)
				self.ref.astype(float)
			except ValueError:
				pass # just ignore invalid arms
		if not isinstance(self.sam, np.ndarray):
			try:
				self.sam = np.array(self.sam)
				self.sam.astype(float)
			except ValueError:
				pass # just ignore invalid arms
		if len(self.ref) == 0:
			self.y_norm = self.y
		else:
			self.y_norm = (self.y - self.ref - self.sam) / (2 * np.sqrt(self.sam * self.ref))
			self._is_normalized = True
		self.plotwidget = plt
		self.xmin = None
		self.xmax = None
		self.probably_wavelength = None
		self._check_domain()


	def _safe_cast(self):
		'''
		Prevent attributes to be modified inplace.
		'''
		x, y, ref, sam = np.copy(self.x), np.copy(self.y), np.copy(self.ref), np.copy(self.sam)
		return x, y, ref, sam


	def _check_domain(self):
		"""
		Checks the domain of data just by looking at x axis' minimal value.
		FIXME: Units are obviously not added yet, we work in nm and PHz...
		"""
		if min(self.x) > 100:
			self.probably_wavelength = True
		else:
			self.probably_wavelength = False

	@classmethod
	def parse_raw(cls, basefile, ref=None, sam=None, skiprows=8,
		decimal=',', sep=';', meta_len=4):
		'''
		Dataset object alternative constructor. Helps to load in data just by giving the filenames
		in the target directory.

		Parameters:
		----------
		basefile: str
			base interferogram
			*.trt file generated by the spectrometer

		ref: str, optional
			reference arm's spectra
			*.trt file generated by the spectrometer

		sam: str, optional
			sample arm's spectra
			*.trt file generated by the spectrometer

		skiprows: int, optional
			Skip rows at the top of the file. Default is 8.

		sep: string, optional
			The delimiter in the original interferogram file.
			Default is ';'.

		decimal: string, optional
			Character recognized as decimal separator in the original dataset. 
			Often ',' for European data.
			Default is ','.

		meta_len: int, optional
			The first n lines in the original file containing the meta information
			about the dataset. Default is 4.
		'''
		with open(basefile) as file:
			cls._metadata = ''.join(next(file) for _ in range(meta_len))
		df = pd.read_csv(basefile, skiprows=skiprows, sep=sep, decimal=decimal, names=['x', 'y'])
		if ref is not None and sam is not None:
			r = pd.read_csv(ref, skiprows=skiprows, sep=sep, decimal=decimal, names=['x', 'y'])
			s = pd.read_csv(sam, skiprows=skiprows, sep=sep, decimal=decimal, names=['x', 'y'])
			return cls(df['x'].values, df['y'].values, r['y'].values, s['y'].values)
		return cls(df['x'].values, df['y'].values)

	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		string = f'''
{type(self).__name__} object

Parameters
----------
Datapoints = {len(self.x)}
Normalized: {self._is_normalized}
Arms are separated: {True if len(self.ref) > 0 else False}
Predicted domain: {'wavelength' if self.probably_wavelength else 'frequency'}

Metadata extracted from file
----------------------------
{str(self._metadata)}
		'''
		return string

	@property
	def data(self):
		"""
		Returns the *current* dataset as pandas DataFrame.
		"""
		if self._is_normalized:
			try:
				self._data = pd.DataFrame({
				'x': self.x,
				'y': self.y,
				'sample': self.sam,
				'reference': self.ref,
				'y_normalized': self.y_norm
					})
			except ValueError:
				self._data = pd.DataFrame({
					'x': self.x,
					'y': self.y,
					})
		else:
			self._data = pd.DataFrame({
				'x': self.x,
				'y': self.y,
				})
		return self._data

	@property
	def is_normalized(self):
		"""
		Retuns whether the dataset is normalized.
		"""
		return self._is_normalized
	
	def chdomain(self):
		""" Changes from wavelength [nm] to ang. freq. [PHz] domain and vica versa."""
		self.x = (2*np.pi*C_LIGHT)/self.x
		self._check_domain()

	def detect_peak_cwt(self, width, floor_thres=0.05):
		x, y, ref, sam = self._safe_cast()
		xmax, ymax, xmin, ymin = cwt(x, y, ref, sam, width=width, floor_thres=floor_thres)
		return xmax, ymax, xmin, ymin

	def savgol_fil(self, window=10, order=3):
		"""
		Applies Savitzky-Golay filter on the dataset.

		Parameters:
		----------
		window: int, default is 10
			length of the convolutional window for the filter

		order: int, default is 3
			Degree of polynomial to fit after the convolution.
			If not odd, it's incremented by 1. Must be lower than window.
			Usually it's a good idea to stay with a low degree, e.g 3 or 5.

		Notes:
		------

		If arms were given, it will merge them into the self.y and self.y_norm variables.
		Also applies a linear interpolation on dataset (and raises warning).
		"""
		self.x, self.y_norm = savgol(self.x, self.y, self.ref, self.sam, window=window, order=order)
		self.y = self.y_norm
		self.ref = []
		self.sam = []
		warnings.warn('Linear interpolation have been applied to data.', InterpolationWarning)
		
	def slice(self, start=-9999, stop=9999):
		"""
		Cuts the dataset on x axis in this form: [start, stop]

		Parameters:
		----------
		start: float, default is -9999
			start value of cutting interval
			Not giving a value will keep the dataset's original minimum value.
			Note that giving -9999 will leave original minimum untouched too.

		stop: float, default is 9999
			stop value of cutting interval
			Not giving a value will keep the dataset's original maximum value.
			Note that giving 9999 will leave original maximum untouched too.

		Notes:
		------

		If arms were given, it will merge them into the self.y and self.y_norm variables.
		"""
		self.x, self.y_norm = cut_data(self.x, self.y, self.ref, self.sam, startValue=start, endValue=stop)
		self.ref = []
		self.sam = []
		self.y = self.y_norm
		# Just to make sure it's correctly shaped. Later on we might delete this.
		if type(self).__name__ == 'FFTMethod':
			self.original_x = self.x

	def convolution(self, window_length, std=20):
		"""
		Applies a convolution with a gaussian on the dataset

		Parameters:
		----------
		window_length: int
			length of the gaussian window

		std: float, default is 20
			standard deviation of the gaussian

		Notes:
		------

		If arms were given, it will merge them into the self.y and self.y_norm variables.
		Also applies a linear interpolation on dataset (and raises warning).
		"""
		self.x, self.y_norm = convolution(self.x, self.y, self.ref, self.sam, window_length, standev=std)
		self.ref = []
		self.sam = []
		self.y = self.y_norm
		warnings.warn('Linear interpolation have been applied to data.', InterpolationWarning)


	def detect_peak(self, pmax=0.1, pmin=0.1, threshold=0.1, except_around=None):
		"""
		Basic algorithm to find extremal points in data.

		Parameters:
		----------

		pmax: float, default is 0.1
			prominence of maximum points
			the lower it is, the more peaks will be found

		pmin: float, default is 0.1
			prominence of minimum points
			the lower it is, the more peaks will be found

		threshold: float, default is 0
			sets the minimum distance (measured on y axis) required for a point to be
			accepted as extremal

		except_around: interval (array or tuple), default is None
			Overwrites the threshold to be 0 at the given interval.
			format is (lower, higher) or [lower, higher].

		Returns:
		-------
		xmax: array-like
			x coordinates of the maximums

		ymax: array-like
			y coordinates of the maximums

		xmin: array-like
			x coordinates of the minimums

		ymin: array-like
			y coordinates of the minimums
		"""
		x, y, ref, sam = self._safe_cast()
		xmax, ymax, xmin, ymin = find_peak(x, y, ref, sam, proMax=pmax, proMin=pmin, threshold=threshold, except_around=except_around)
		return xmax, ymax, xmin, ymin
		
	def show(self):
		"""
		Draws a graph of the current dataset using matplotlib.
		"""
		if np.iscomplexobj(self.y):
			self.plotwidget.plot(self.x, np.abs(self.y))
		else:   
			try:
				self.plotwidget.plot(self.x, self.y_norm, 'r')
			except Exception:
				self.plotwidget.plot(self.x, self.y, 'r')
		self.plotwidget.grid()
		self.plotwidget.show()


	def normalize(self, filename=None):
		'''
		Normalize the interferogram by finding upper and lower envelope
		on an interactive matplotlib editor.

		Parameters
		----------

		filename: str, default None
			Save the normalized interferogram named by filename in the 
			working directory. If not given it will not be saved.

		Returns
		-------
		None
		'''
		if run_from_ipython():
			return '''It seems you run this code in IPython. Interactive plotting is not yet supported. Consider running it in the regular console.'''
		_l_env = DraggableEnvelope(self.x, self.y, 'l')
		y_transform = _l_env.get_data()
		_u_env = DraggableEnvelope(self.x, y_transform, 'u')
		y_final = _u_env.get_data()
		self.y = y_final
		self.y_norm = y_final
		self._is_normalized = True
		self.plotwidget.title('Final')
		self.show()
		if filename:
			if not filename.endswith('.txt'):
				filename += '.txt'
			np.savetxt(filename, np.transpose([self.x, self.y]), delimiter=',')
			print(f'Successfully saved as {filename}')


class MinMaxMethod(Dataset):
	"""
	Basic interface for Minimum-Maximum Method.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

	# TODO: fix docstring
	def init_edit_session(
		self, pmax=0.1, pmin=0.1, threshold=0, 
		except_around=None, engine='normal', width=10
		):
		""" Function to initialize peak editing on a plot.
		Right clicks will delete the closest point, left clicks 
		will add a new point. Just close the window when finished.
		
		Parameters:
		----------

		pmax: float, default is 0.1
			prominence of maximum points
			the lower it is, the more peaks will be found

		pmin: float, default is 0.1
			prominence of minimum points
			the lower it is, the more peaks will be found

		threshold: float, default is 0
			sets the minimum distance (measured on y axis) required for a point to be
			accepted as extremal

		except_around: interval (array or tuple), default is None
			Overwrites the threshold to be 0 at the given interval.
			format is (lower, higher) or [lower, higher].

		Notes:
		------

		Currently this function is disabled when running it from IPython.
		"""
		if run_from_ipython():
			return '''It seems you run this code in IPython. Interactive plotting is not yet supported. Consider running it in the regular console.'''
		engines = ('cwt', 'normal', 'slope')
		if engine not in engines:
			raise ValueError(f'Engine must be in {str(engines)}')
		if engine == 'normal':
			_x, _y, _xx, _yy = self.detect_peak(
			pmax=pmax, pmin=pmin, threshold=threshold, except_around=except_around
			)

		elif engine == 'slope':
			x, _, _, _ = self._safe_cast()
			y = np.copy(self.y_norm)
			if _maybe_increase_before_cwt(y):
				y += 2
			_, lp, lloc = calc_envelope(y, np.arange(len(y)), 'l')
			_, up, uloc = calc_envelope(y, np.arange(len(y)), 'u')
			lp -= 2
			up -= 2
			_x, _xx = x[lloc], x[uloc]
			_y, _yy = lp, up

		elif engine == 'cwt':
			_x, _y, _xx, _yy = self.detect_peak_cwt(width)
		_xm = np.append(_x, _xx)
		_ym = np.append(_y, _yy)

		try:
			_editpeak = EditPeak(self.x, self.y_norm, _xm, _ym)
		except ValueError:
			_editpeak = EditPeak(self.x, self.y, _xm, _ym)
		# automatically propagate these points to the mins and maxes
		# just in case the default argrelextrema is definitely not called in evaluate.py/min_max_method:
		self.xmin = _editpeak.get_dat[0][:len(_editpeak.get_dat[0])//2]
		self.xmax = _editpeak.get_dat[0][len(_editpeak.get_dat[0])//2:] 
		print(f'In total {len(_editpeak.get_dat[0])} extremal points were recorded.')
		return _editpeak.get_dat[0]

	@print_disp
	def calculate(self, reference_point, order, show_graph=False):
		""" 
		MinMaxMethod's calculate function.

		Parameters:
		----------

		reference_point: float
			reference point on x axis

		fit_order: int
			Polynomial (and maximum dispersion) order to fit. Must be in [1,5].

		show_graph: bool
			shows a the final graph of the spectral phase and fitted curve.

		Returns:
		-------

		dispersion: array-like
			[GD, GDD, TOD, FOD, QOD]

		dispersion_std: array-like
			standard deviations due to uncertanity of the fit
			[GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

		fit_report: lmfit report
			if lmfit is available, the fit report


		Notes:
		------

		Decorated with print_disp, so the results are immediately printed without explicitly saying so.
		"""
		dispersion, dispersion_std, fit_report = min_max_method(
			self.x, self.y, self.ref, self.sam, ref_point=reference_point,
			maxx=self.xmax, minx=self.xmin, fitOrder=order, showGraph=show_graph
			)
		return dispersion, dispersion_std, fit_report


class CosFitMethod(Dataset):
	"""
	Basic interface for the Cosine Function Fit Method.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.params = [1, 1, 1, 1, 1, 1, 1, 1]
		self.fit = None
		self.mt = 8000
		self.f = None


		# FIXME : DIFFERENT ERROR CASES SHOULD BE TAKEN CARE OF
	def predict(self, reference_point=2.355, pmax=0.5, pmin=0.5, threshold=0.35):
		x_min, _, x_max, _ = self.detect_peak(pmax=pmax, pmin=pmin, threshold=threshold)
		try:
			closest_val, idx1 = find_nearest(x_min, reference_point)
			m_closest_val, m_idx1 = find_nearest(x_max, reference_point)
		except ValueError:
			print('Prediction failed.\nSkipping.. ')
			return
		truncated = np.delete(x_min, idx1)
		second_closest_val, _ = find_nearest(truncated, reference_point)
		m_truncated = np.delete(x_max, m_idx1)
		m_second_closest_val, _ = find_nearest(m_truncated, reference_point)
		lowguess = 2*np.pi/np.abs(closest_val-second_closest_val)
		highguess = 2*np.pi/np.abs(m_closest_val-m_second_closest_val)
		self.params[3] = (lowguess+highguess)/2
		print(f'The predicted GD is ± {((lowguess+highguess)/2):.5f} based on reference point of {reference_point}.')

	def set_max_tries(self, value):
		"""
		Overwrite the default scipy maximum try setting to fit the curve.
		"""
		self.mt = value

	def adjust_offset(self, value):
		"""
		Initial guess for offset
		"""
		self.params[0] = value

	def adjust_amplitude(self, value):
		"""
		Initial guess for amplitude
		"""
		self.params[1] = value

	def guess_GD(self, value):
		"""
		Initial guess for GD in fs
		"""
		self.params[3] = value

	def guess_GDD(self, value):
		"""
		Initial guess for GDD in fs^2
		"""
		self.params[4] = value / 2

	def guess_TOD(self, value):
		"""
		Initial guess for TOD in fs^3
		"""
		self.params[5] = value / 6

	def guess_FOD(self, value):
		"""
		Initial guess for FOD in fs^4
		"""
		self.params[6] = value / 24

	def guess_QOD(self, value):
		"""
		Initial guess for QOD in fs^5
		"""
		self.params[7] = value / 120

	def set_max_order(self, order):
		"""
		Sets the maximum order of dispersion to look for.
		Should be called after guessing the initial parameters.

		Parameters:
		----------

		order: int
			maximum order of dispersion to look for. Must be in [1, 5]
		"""
		if order > 5 or order < 1:
			print(f'Order should be an in integer from [1,5], currently {order} is given')
		try:
			int(order)
		except ValueError:
			print(f'Order should be an in integer from [1,5], currently {order} is given')
		order = 6 - order
		for i in range(1, order):
			self.params[-i] = 0

	@print_disp
	def calculate(self, reference_point):
		""" 
		Cosine fit's calculate function.

		Parameters:
		----------
		reference_point: float
			reference point on x axis

		Returns:
		-------

		dispersion: array-like
			[GD, GDD, TOD, FOD, QOD]

		dispersion_std: array-like
			standard deviations due to uncertanity of the fit
			[GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

		fit_report: lmfit report
			WILL BE IMPLEMENTED


		Notes:
		------

		Decorated with print_disp, so the results are immediately printed without explicitly saying so.
		"""
		dispersion, self.fit = cff_method(self.x, self.y, self.ref, self.sam, 
			ref_point=reference_point, p0=self.params, maxtries=self.mt)
		dispersion = list(dispersion)
		while len(dispersion) < 5:
			dispersion.append(0)
		return dispersion, [0, 0, 0, 0, 0], 'Fit report for CFF not supported yet.'

	def plot_result(self):
		"""
		If the fitting happened, draws the fitted curve on the original dataset.
		Also prints the coeffitient of determination of the fit (a.k.a. r^2).
		"""
		try:
			residuals = self.y_norm - self.fit
			ss_res = np.sum(residuals**2)
			ss_tot = np.sum((self.y_norm - np.mean(self.y_norm))**2)
			print('r^2 = ' + str(1 - (ss_res / ss_tot)))
		except Exception:
			pass
		if self.fit is not None:
			self.plotwidget.plot(self.x, self.fit, 'k--', label='fit', zorder=99)
			self.plotwidget.legend()
			self.show()
		else:
			self.show()

	def optimizer(self, reference_point, order=3, initial_region_ratio=0.1,
		extend_by=0.1, coef_threshold=0.3, max_tries=5000, show_endpoint=True):
		"""
		Cosine fit optimizer. It's based on adding new terms to fit function successively
		until we reach the max_order.

		Notes
		-----
		If the fit fails some parameters must be tweaked in order to achieve results.
		There is a list below with issues, its suspected reasons and solutions.

		**SciPy raises OptimizeWarning and the affected area is small or not showing
		  any fit

		Reasons:
		- Completely wrong initial GD guess (or lack of guessing).
		- Too broad inital region, so that the optimizer cannot find a suitable fit.
		  This usually happens when the used data is large, or the spectral resolution
		  is high.

		Solution:
		- Provide better inital guess for GD.
		- Lower the inital_region_ratio.

		**SciPy raises OptimizeWarning and the affected area is bigger

		Reasons:
		- When the optimizer steps up with order it also extends the region of fit.
		This error usually present when the region of fit is too quickly growing.

		Solution:
		- Lower extend_by argument.

		**The optimizer is finished, but wrong fit is produced.

		Reasons:
		- We measure the goodness of fit with r^2 value. To allow this
		optimizer to smoothly find appropriate fits even for noisy datasets
		it's a good practice to keep the r^2 a lower value, such as the default 0.3.
		The way it works is we step up in order of fit (until max order) and extend
		region every time when a fit reaches the specified r^2 threshold value.
		This can be controlled via the coef_threshold argument.

		Solution:
		- Adjust the coef_threshold value. Note that it's highly recommended not to
		set a higher value than 0.6.
		"""

		x, y, ref, sam = self._safe_cast()
		self.f = FitOptimizer(x, y, ref, sam, reference_point=reference_point,
		max_order=order)
		self.f.set_initial_region(initial_region_ratio)
		self.f.set_final_guess(GD=self.params[3], GDD=self.params[4], TOD=self.params[5],
		FOD=self.params[6], QOD=self.params[7]) # we can pass it higher params safely, they are ignored.
		self.f.run_loop(extend_by, coef_threshold, max_tries=max_tries, show_endpoint=show_endpoint)

class SPPMethod(Dataset):
	"""
	Basic interface for Stationary Phase Point Method. It will be improved later.
	"""
	raw = False

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.om = None
		self.de = None
		self.bf = None

	@classmethod
	def from_raw(cls, omegas, delays):
		"""
		Alternative constructor to work with matching pairs of data.

		Parameters:
		----------

		omegas: array-like
			the angular frequency values at SPP positions

		delays: array-like
			the delay of the corresponding omega 

		"""
		cls.raw = True
		return cls(omegas, delays)

	@print_disp
	def calculate(self, reference_point, order):
		""" 
		SPP's calculate function.

		Parameters:
		----------

		reference_point: float
			reference point on x axis

		fit_order: int
			Polynomial (and maximum dispersion) order to fit. Must be in [1,4].

		Returns:
		-------

		dispersion: array-like
			[GD, GDD, TOD, FOD, QOD]

		dispersion_std: array-like
			standard deviations due to uncertanity of the fit
			[GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

		bf: array-like
			if lmfit is available, the best fitting curve

		Notes:
		------

		Decorated with print_disp, so the results are immediately printed without explicitly saying so.
		"""
		if self.raw:
			_, _, dispersion, dispersion_std, self.bf = spp_method(
				self.y, self.x, reference_point=reference_point, fitOrder=order, from_raw=True
				)
			self.om = self.x
			self.de = self.y
		else:
			self.om, self.de, dispersion, dispersion_std, self.bf = spp_method(
				self.y, self.x, fitOrder=order, from_raw=False
				)
		dispersion = list(dispersion)
		dispersion_std = list(dispersion_std)
		while len(dispersion)<5:
			dispersion.append(0)
		while len(dispersion_std)<5:
			dispersion_std.append(0)
		return dispersion, dispersion_std, self.bf

	def plot_result(self):
		"""
		If the fitting happened, draws the fitted curve on the original dataset.
		"""
		self.plotwidget.plot(self.om, self.de, 'o')
		try:
			self.plotwidget.plot(self.om, self.bf, 'r--', zorder=1)
		except Exception:
			pass
		self.plotwidget.show()


class FFTMethod(Dataset):
	"""
	Basic interface for the Fourier transform method.
	# FIXME: bug when calling show_graph on calulcate method: it opens up a clean figure.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		#making sure it's not normalized
		if self._is_normalized:
			self.y_norm = self.y
			self._is_normalized = False
		self.original_x = self.x
		self.at = None
		self.std = None
		self.window_order = None
		self._ifft_called_first = False

	def shift(self, axis='x'):
		"""
		Equivalent to scipy.fftpack.fftshift, but it's easier to
		use this function instead, because we don't need to explicitly
		call the class' x and y attribute.
		
		Parameter(s):
		------------
		axis: str, default is 'x'
			either 'x', 'y' or 'both'
		"""
		if axis == 'x':
			self.x = fftshift(self.x)
		elif axis == 'y':
			self.y = fftshift(self.y)
		elif axis == 'both':
			self.y = fftshift(self.y)
			self.x = fftshift(self.x)
		else:
			raise ValueError(f'axis should be either x, y or both, currently {axis} is given.')

	def ifft(self, interpolate=True):
		"""
		Applies ifft to the dataset.
		
		Parameter(s):
		------------

		interpolate: bool, default is True
			Whether to apply linear interpolation on the dataset before transforming.
		"""
		self._ifft_called_first = True
		self.x, self.y = ifft_method(self.x, self.y, interpolate=interpolate)

	def fft(self):
		"""
		Applies fft to the dataset.
		If ifft was not called first, inaccurate results might happen. It will be fixed later on.
		Check calculate function's docstring for more detail.
		"""
		if not self._ifft_called_first:
			warnings.warn('This module is designed to call ifft before fft, so inconsistencies might occur when calling fft first. Consider using numpys fft package with your own logic. This functionality will be added later on.', FourierWarning)
		self.x, self.y = fft_method(self.original_x, self.y)

	def window(self, at, std, window_order=6):
		"""
		Draws a gaussian window on the plot with the desired parameters.
		The maximum value is adjusted for the dataset mostly for visibility reasons.
		You should explicitly call self.show() after this function is set.

		Parameters:
		----------

		at: float
			maximum of the gaussian curve

		std: float #FIXME: RENAME THIS TO FWHM
			Full width at half maximum of the gaussian

		window_order: int, default is 6
			Order of the gaussian curve.
			If not even, it's incremented by 1 for safety reasons.
		"""
		self.at = at
		self.std = std
		self.window_order = window_order
		gaussian = gaussian_window(self.x, self.at, self.std, self.window_order)
		self.plotwidget.plot(self.x, gaussian*max(abs(self.y)), 'r--')

	def apply_window(self):
		"""
		If window function is correctly set, applies changes to the dataset.
		"""
		self.plotwidget.clf()
		self.plotwidget.cla()
		self.plotwidget.close()
		self.y = cut_gaussian(self.x, self.y, spike=self.at, sigma=self.std, win_order=self.window_order)
		
	@print_disp
	def calculate(self, reference_point, fit_order, show_graph=False):
		""" 
		FFTMethod's calculate function.

		Parameters:
		----------

		reference_point: float
			reference point on x axis

		fit_order: int
			Polynomial (and maximum dispersion) order to fit. Must be in [1,5].

		show_graph: bool, optional
			shows a the final graph of the spectral phase and fitted curve.

		Returns:
		-------

		dispersion: array-like
			[GD, GDD, TOD, FOD, QOD]

		dispersion_std: array-like
			standard deviations due to uncertanity of the fit
			[GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

		fit_report: lmfit report
			if lmfit is available, the fit report

		Notes:
		------

		Decorated with print_disp, so the results are immediately printed without explicitly saying so.

		Currently the x-axis transformation is sloppy, because we cache the original x axis and not transforming it
		backwards. In addition we need to keep track of interpolation and zero-padding too.
		Currently the transforms are correct only if first ifft was used.
		For now it's doing okay: giving good results. 
		For consistency we should still implement that a better way later.
		"""
		dispersion, dispersion_std, fit_report = args_comp(
			self.x, self.y, reference_point=reference_point, fitOrder=fit_order, showGraph=show_graph
			)
		return dispersion, dispersion_std, fit_report



''' Testing the main window base functionality. '''
import sys
import unittest
import numpy as np
from PyQt5.QtWidgets import QApplication
from PyQt5.QtTest import QTest
from PyQt5.QtCore import Qt
from pysprint.logic import SPPWindow


class TestSPP(unittest.TestCase):
    pass
    # def setUp(self):
    #     self.app = QApplication(sys.argv)
    #     self.spp = SPPWindow()
    #     self.spp.show()

    # def tearDown(self):
    #     self.spp.close()

    # def test_clean(self):
    #     self.spp.clean_up()
    #     self.spp.preview_data()
    #     self.spp.record_delay()


if __name__ == "__main__":
    unittest.main()



from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from PyQt5.QtWidgets import QWidget, QVBoxLayout

class PlotWidget(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.canvas = FigureCanvas(Figure())
        vertical_layout = QVBoxLayout()
        vertical_layout.addWidget(self.canvas)
        self.canvas.axes = self.canvas.figure.add_subplot(111)
        self.setLayout(vertical_layout)


import warnings

warnings.filterwarnings("ignore", category=UserWarning) # for matplotlib

import numpy as np
import matplotlib
import matplotlib.pyplot as plt 
from matplotlib.backend_bases import MouseButton

# ---------------------------MONKEY PATCH-------------------------------------
import matplotlib.cbook as cbook
import matplotlib.widgets as widgets
from matplotlib.backend_tools import ToolBase
from matplotlib.backend_managers import ToolManager

def tm_init(self, figure=None):
    self._key_press_handler_id = None

    self._tools = {}
    self._keys = {}
    self._toggled = {}
    self._callbacks = cbook.CallbackRegistry()

    # to process keypress event
    self.keypresslock = widgets.LockDraw()
    self.messagelock = widgets.LockDraw()

    self._figure = None
    self.set_figure(figure)

matplotlib.backend_managers.ToolManager.__init__ = tm_init

def new_init(self, toolmanager, name):
        self._name = name
        self._toolmanager = toolmanager
        self._figure = None

matplotlib.backend_tools.ToolBase.__init__ = new_init

# ----------------------------------------------------------------------------


from matplotlib.backend_tools import ToolToggleBase
from pysprint.utils import get_closest
# TODO: implement real euclidean metrics with epsilon distance



class SelectButton(ToolToggleBase):
    """
    Toggle button on matplotlib toolbar.
    """
    description = 'Enable click records'
    default_toggled = True
    default_keymap = 't'
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class EditPeak(object):
	"""
	This class helps to record and delete peaks from a dataset.
	Right clicks will delete the closest (distance is measured with regards to x)
	extremal point found on the graph, left clicks will add a new point.
	Edits can be saved by just closing the matplotlib window.
	Returns the x coordinates of the selected points.
	Note that this class shouldn't be explicitly called by the user.
	"""
	def __init__(self, x, y, x_extremal=None, y_extremal=None):

		# This is here because we do not want other figures
		# to be affected by this.
		matplotlib.rcParams["toolbar"] = "toolmanager"
		
		self.figure = plt.figure()
		self.cid = None
		self.x = x
		self.y = y
		plt.plot(self.x, self.y, 'r')
		self.x_extremal = x_extremal
		self.y_extremal = y_extremal
		if not len(self.x_extremal) == len(self.y_extremal):
			raise ValueError('Data shapes are different')
		self.press()
		self.lins, = plt.plot(
			self.x_extremal, self.y_extremal, 'ko', markersize=6, zorder=99
			)
		plt.grid(alpha=0.7)
		# adding the button to navigation toolbar
		tm = self.figure.canvas.manager.toolmanager
		tm.add_tool('Toggle recording', SelectButton)
		self.figure.canvas.manager.toolbar.add_tool(
			tm.get_tool('Toggle recording'), "toolgroup"
			)
		self.my_select_button = tm.get_tool('Toggle recording')
		plt.show()

	def on_clicked(self, event):
	        """ Function to record and discard points on plot."""
	        ix, iy = event.xdata, event.ydata
	        if event.inaxes is None:
	        	return
	        if self.my_select_button.toggled:
		        if event.button is MouseButton.RIGHT:
		        	ix, iy, idx = get_closest(ix, self.x_extremal, self.y_extremal)
		        	self.x_extremal = np.delete(self.x_extremal, idx)
		        	self.y_extremal = np.delete(self.y_extremal, idx)
		        elif event.button is MouseButton.LEFT:
		        	ix, iy, idx = get_closest(ix, self.x, self.y)
		        	self.x_extremal = np.append(self.x_extremal, ix)
		        	self.y_extremal = np.append(self.y_extremal, iy)
		       	else:
		       		pass
		        self.lins.set_data(self.x_extremal, self.y_extremal)
		        plt.draw()
		        return

	def press(self):
	        """Usual function to connect matplotlib.."""
	        self.cid = self.figure.canvas.mpl_connect(
	        	'button_press_event', self.on_clicked
	        	)

	def release(self):
	        """ On release functionality. It's never called but we will
	        need this later on.."""
	        self.figure.canvas.mpl_disconnect(self.cid)

	@property
	def get_dat(self):
		""" Returns the the selected points."""
		return self.lins.get_data()



# -*- coding: utf-8 -*-
from math import factorial
import operator

import numpy as np
import matplotlib.pyplot as plt
import scipy
from scipy.optimize import curve_fit
from scipy.signal import argrelextrema

try:
    from lmfit import Model
    _has_lmfit = True
except ImportError:
    _has_lmfit = False


from pysprint.utils import (
    findNearest, _handle_input, lmfit_disp, fourier_interpolate
    )


__all__ = [
    'min_max_method', 'cos_fit1', 'cos_fit2', 'cos_fit3',
    'cos_fit4', 'cos_fit5', 'spp_method', 'cff_method', 'fft_method',
    'cut_gaussian', 'ifft_method', 'args_comp'
    ]


def min_max_method(
        initSpectrumX, initSpectrumY, referenceArmY, sampleArmY, ref_point,
        maxx=None, minx=None, fitOrder=5, showGraph=False
        ):
    """Calculates the dispersion with minimum-maximum method

    Parameters
    ----------

    initSpectrumX: array-like
    x-axis data

    initSpectrumY: array-like
    y-axis data

    referenceArmY, sampleArmY: array-like
    reference and sample arm spectra evaluated at initSpectrumX

    ref_point: float
    the reference point to calculate order

    maxx and minx: array-like, optional
    the accepted minimal and maximal x values (if you want to manually pass)

    fitOrder: int, optional
    degree of polynomial to fit data [1, 5]

    showGraph: bool, optional
    if True returns a matplotlib plot and pauses execution until closing the window

    Returns
    -------

    dispersion: array-like
    [GD, GDD, TOD, FOD, QOD]

    dispersion_std: array-like
    [GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

    fit_report: lmfit report object
    """
    Xdata, Ydata = _handle_input(
        initSpectrumX, initSpectrumY, referenceArmY, sampleArmY
        )
    if maxx is None:
        maxInd = argrelextrema(Ydata, np.greater)
        maxx = Xdata[maxInd]
    if minx is None:
        minInd = argrelextrema(Ydata, np.less)
        minx = Xdata[minInd]

    _, ref_index = findNearest(Xdata, ref_point)

    relNegMaxFreqs = np.array([a for a in (Xdata[ref_index] - maxx) if a < 0])
    relNegMinFreqs = np.array([b for b in (Xdata[ref_index] - minx) if b < 0])
    relNegFreqs = sorted(np.append(relNegMaxFreqs, relNegMinFreqs))
    relNegFreqs = relNegFreqs[::-1]
    relPosMaxFreqs = np.array([c for c in (Xdata[ref_index] - maxx) if c > 0])
    relPosMinFreqs = np.array([d for d in (Xdata[ref_index] - minx) if d > 0])
    relPosFreqs = sorted(np.append(relPosMinFreqs, relPosMaxFreqs))

    if len(relNegFreqs) == 0 and len(relPosFreqs) == 0:
        raise ValueError('No extremal points found.')
    negValues = np.zeros_like(relNegFreqs)
    posValues = np.zeros_like(relPosFreqs)
    for freq in range(len(relPosFreqs)):
        posValues[freq] = np.pi * (freq + 1)
    for freq in range(len(relNegFreqs)):
        negValues[freq] = np.pi * (freq + 1)
    x_s = np.append(relPosFreqs, relNegFreqs)
    y_s = np.append(posValues, negValues)

    # FIXME: Do we even need this?
    idx = np.argsort(x_s)
    fullXValues, fullYValues = x_s[idx], y_s[idx]
    
    if _has_lmfit:
        if fitOrder == 5:
            fitModel = Model(polynomialFit5)
            params = fitModel.make_params(b0=0, b1=1, b2=1, b3=1, b4=1, b5=1)
            result = fitModel.fit(fullYValues, x=fullXValues, params=params, method='leastsq')
        elif fitOrder == 4:
            fitModel = Model(polynomialFit4)
            params = fitModel.make_params(b0=0, b1=1, b2=1, b3=1, b4=1)
            result = fitModel.fit(fullYValues, x=fullXValues, params=params, method='leastsq')
        elif fitOrder == 3:
            fitModel = Model(polynomialFit3)
            params = fitModel.make_params(b0=0, b1=1, b2=1, b3=1)
            result = fitModel.fit(fullYValues, x=fullXValues, params=params, method='leastsq')
        elif fitOrder == 2:
            fitModel = Model(polynomialFit2)
            params = fitModel.make_params(b0=0, b1=1, b2=1)
            result = fitModel.fit(fullYValues, x=fullXValues, params=params, method='leastsq')
        elif fitOrder == 1:
            fitModel = Model(polynomialFit1)
            params = fitModel.make_params(b0=0, b1=1)
            result = fitModel.fit(fullYValues, x=fullXValues, params=params, method='leastsq')
        else:
            raise ValueError('Order is out of range, please select from [1,5]')
    else:
        if fitOrder == 5:
            popt, pcov = curve_fit(polynomialFit5, fullXValues, fullYValues, maxfev=8000)
            _function = polynomialFit5
        elif fitOrder == 4:
            popt, pcov = curve_fit(polynomialFit4, fullXValues, fullYValues, maxfev=8000)
            _function = polynomialFit4
        elif fitOrder == 3:
            popt, pcov = curve_fit(polynomialFit3, fullXValues, fullYValues, maxfev=8000)
            _function = polynomialFit3
        elif fitOrder == 2:
            popt, pcov = curve_fit(polynomialFit2, fullXValues, fullYValues, maxfev=8000)
            _function = polynomialFit2
        elif fitOrder == 1:
            popt, pcov = curve_fit(polynomialFit1, fullXValues, fullYValues, maxfev=8000)
            _function = polynomialFit1
        else:
            raise ValueError('Order is out of range, please select from [1,5]')
    try:
        if _has_lmfit:
            dispersion, dispersion_std = lmfit_disp(result.params.items())
            dispersion = dispersion[1:]
            dispersion_std = dispersion_std[1:]
            for idx in range(len(dispersion)):
                dispersion[idx] = dispersion[idx] * factorial(idx+1)
                dispersion_std[idx] = dispersion_std[idx] * factorial(idx+1)
            while len(dispersion) < 5:
                dispersion.append(0)
                dispersion_std.append(0)
            fit_report = result.fit_report()
        else:
            fullXValues = np.asarray(fullXValues)
            dispersion = []
            dispersion_std = []
            for idx in range(len(popt) - 1):
                dispersion.append(popt[idx + 1] * factorial(idx + 1))
            while len(dispersion) < 5:
                dispersion.append(0)
            while len(dispersion_std) < len(dispersion):
                dispersion_std.append(0)
            fit_report = '\nTo display detailed results, you must have lmfit installed.'
        if showGraph:
            fig = plt.figure(figsize=(7, 7))
            fig.canvas.set_window_title('Min-max method fitted')
            plt.plot(fullXValues, fullYValues, 'o', label='dataset')
            try:
                plt.plot(fullXValues, result.best_fit, 'r*', label='fitted')
            except Exception:
                plt.plot(fullXValues, _function(fullXValues, *popt), 'r*', label='fitted')
            plt.legend()
            plt.grid()
            plt.show()
        return dispersion, dispersion_std, fit_report
    except Exception as e:
        return [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], e


def polynomialFit5(x, b0, b1, b2, b3, b4, b5):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6
    b4 = FOD / 24
    b5 = QOD / 120
    """
    return b0+b1*x+b2*x**2+b3*x**3+b4*x**4+b5*x**5


def polynomialFit4(x, b0, b1, b2, b3, b4):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6
    b4 = FOD / 24
    """
    return b0+b1*x+b2*x**2+b3*x**3+b4*x**4


def polynomialFit3(x, b0, b1, b2, b3):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6

    """
    return b0+b1*x+b2*x**2+b3*x**3


def polynomialFit2(x, b0, b1, b2):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    """
    return b0+b1*x+b2*x**2


def polynomialFit1(x, b0, b1):
    """
    Taylor polynomial for fit
    b1 = GD
    """
    return b0+b1*x


def cos_fit1(x, c0, c1, b0, b1):
    return c0 + c1 * np.cos(polynomialFit1(x, b0, b1))


def cos_fit2(x, c0, c1, b0, b1, b2):
    return c0 + c1 * np.cos(polynomialFit2(x, b0, b1, b2))


def cos_fit3(x, c0, c1, b0, b1, b2, b3):
    return c0 + c1 * np.cos(polynomialFit3(x, b0, b1, b2, b3))


def cos_fit4(x, c0, c1, b0, b1, b2, b3, b4):
    return c0 + c1 * np.cos(polynomialFit4(x, b0, b1, b2, b3, b4))


def cos_fit5(x, c0, c1, b0, b1, b2, b3, b4, b5):
    return c0 + c1 * np.cos(polynomialFit5(x, b0, b1, b2, b3, b4, b5))


def spp_method(delays, omegas, reference_point=0, fitOrder=4, from_raw=False):
    """
    Calculates the dispersion from SPP's positions and delays.

    Parameters
    ----------

    delays: array-like
    the time delays in fs
    if from_raw is enabled you must pass matching pairs with omegas

    omegas: array-like
    in form of [[SPP1, SPP2, SPP3, SPP4],[SPP1, SPP2, SPP3, SPP4], ..]
    for lesser SPP cases replace elements with None:
    [[SPP1, None, None, None],[SPP1, None, None, None], ..]
    if from_raw is enabled, you must pass matching pairs with delays

    fitOrder: int
    order of polynomial to fit the given data

    from_raw: bool
    if True you can pass matching pairs to delays and omegas, and it will perform
    a normal curve fitting. It's useful at the API.

    Returns
    -------
    omegas_unpacked: array-like
    x axis data

    delays_unpacked : array-like
    y axis data

    dispersion: array-like
    [GD, GDD, TOD, FOD, QOD]

    dispersion_std: array-like
    [GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

    bf: array-like
    best fitting curve for plotting
    """
    if from_raw:
        delays_unpacked = delays
        omegas_unpacked = omegas
    else:
        delays = delays[delays != np.array(None)]
        omegas_unpacked = []
        delays_unpacked = []
        for delay, element in zip(delays, omegas):
            item = [x for x in element if x is not None]
            omegas_unpacked.extend(item)
            delays_unpacked.extend(len(item) * [delay])
    # FIXME: should be numpy arrays..
    L = sorted(zip(omegas_unpacked, delays_unpacked), key=operator.itemgetter(0))
    omegas_unpacked, delays_unpacked = zip(*L)
    omegas_unpacked = [val-reference_point for val in omegas_unpacked]
    try:
        if _has_lmfit:
            if fitOrder == 2:
                fitModel = Model(polynomialFit2)
                params = fitModel.make_params(b0=1, b1=1, b2=1)
                result = fitModel.fit(delays_unpacked, x=omegas_unpacked, params=params, method='leastsq')
            elif fitOrder == 3:
                fitModel = Model(polynomialFit3)
                params = fitModel.make_params(b0=1, b1=1, b2=1, b3=1)
                result = fitModel.fit(delays_unpacked, x=omegas_unpacked, params=params, method='leastsq')
            elif fitOrder == 4:
                fitModel = Model(polynomialFit4)
                params = fitModel.make_params(b0=1, b1=1, b2=1, b3=1, b4=1)
                result = fitModel.fit(delays_unpacked, x=omegas_unpacked, params=params, method='leastsq')
            elif fitOrder == 1:
                fitModel = Model(polynomialFit1)
                params = fitModel.make_params(b0=1, b1=1)
                result = fitModel.fit(delays_unpacked, x=omegas_unpacked, params=params, method='leastsq')
            else:
                raise ValueError('Order is out of range, please select from [1,4]')
            dispersion, dispersion_std = lmfit_disp(result.params.items())
            for idx in range(len(dispersion)):
                dispersion[idx] = dispersion[idx]*factorial(idx)
            for idx in range(len(dispersion_std)):
                if dispersion_std[idx] is not None:
                    dispersion_std[idx] = dispersion_std[idx] * factorial(idx)
                else:
                    dispersion_std[idx] = 0
            while len(dispersion) < 5:
                dispersion.append(0)
                dispersion_std.append(0)
            while len(dispersion_std) < 5:
                dispersion_std.append(0)
            bf = result.best_fit
        else:
            if fitOrder == 4:
                popt, pcov = curve_fit(polynomialFit4, omegas_unpacked, delays_unpacked, maxfev=8000)
                _function = polynomialFit4
            elif fitOrder == 3:
                popt, pcov = curve_fit(polynomialFit3, omegas_unpacked, delays_unpacked, maxfev=8000)
                _function = polynomialFit3
            elif fitOrder == 2:
                popt, pcov = curve_fit(polynomialFit2, omegas_unpacked, delays_unpacked, maxfev=8000)
                _function = polynomialFit2
            elif fitOrder == 1:
                popt, pcov = curve_fit(polynomialFit1, omegas_unpacked, delays_unpacked, maxfev=8000)
                _function = polynomialFit1
            else:
                raise ValueError('Order is out of range, please select from [1,4]')
            omegas_unpacked = np.asarray(omegas_unpacked)
            dispersion = []
            dispersion_std = []
            for idx in range(len(popt)):
                dispersion.append(popt[idx] * factorial(idx))
            while len(dispersion) < 5:
                dispersion.append(0)
            while len(dispersion_std) < len(dispersion):
                dispersion_std.append(0)
            bf = _function(omegas_unpacked, *popt)
        return omegas_unpacked, delays_unpacked, dispersion, dispersion_std, bf
    except Exception as e:
        return [], [], [e], [], []


def cff_method(
        initSpectrumX, initSpectrumY, referenceArmY, sampleArmY, ref_point=0,
        p0=[1, 1, 1, 1, 1, 1, 1, 1], maxtries=8000
        ):
    """
    Phase modulated cosine function fit method.


    Parameters
    ----------

    initSpectrumX: array-like
    x-axis data

    initSpectrumY: array-like
    y-axis data

    referenceArmY, sampleArmY: array-like
    the reference and sample arm spectra evaluated at initSpectrumX

    p0: array-like
    the initial parameters for fitting

    Returns
    -------

    dispersion: array-like
    [GD, GDD, TOD, FOD, QOD]

    bf: array-like
    best fitting curve

    """
    # TODO: BOUNDS WILL BE SET  ..
    # bounds=((-1000, -10000, -10000, -np.inf, -np.inf, -np.inf, -np.inf, -np.inf),
    # (1000, 10000, 10000, np.inf, np.inf, np.inf, np.inf, np.inf))
    Xdata, Ydata = _handle_input(
        initSpectrumX, initSpectrumY, referenceArmY, sampleArmY
        )
# TODO: replace with lmfit
    try:
        if len(np.trim_zeros(p0, 'b')) + 4 == len(p0):
            _funct = cos_fit1
            p0 = p0[:-4]
        elif p0[-1] == 0 and p0[-2] == 0 and p0[-3] == 0:
            _funct = cos_fit2
            p0 = p0[:-3]
        elif p0[-1] == 0 and p0[-2] == 0:
            _funct = cos_fit3
            p0 = p0[:-2]
        elif p0[-1] == 0:
            _funct = cos_fit4
            p0 = p0[:-1]
        else:
            _funct = cos_fit5
        popt, pcov = curve_fit(_funct, Xdata-ref_point, Ydata, p0, maxfev=maxtries)
        dispersion = np.zeros_like(popt)[:-3]
        for num in range(len(popt)-3):
            dispersion[num] = popt[num+3]*factorial(num+1)
        return dispersion, _funct(Xdata-ref_point, *popt)
    except RuntimeError:
        raise ValueError(f'Max tries ({maxtries}) reached.. \nParameters could not be estimated.')


def fft_method(initSpectrumX, initSpectrumY):
    """Perfoms FFT on data

    Parameters
    ----------

    initSpectrumX: array-like
    the x-axis data

    initSpectrumY: array-like
    the y-axis data

    Returns
    -------
    xf: array-like
    the transformed x data

    yf: array-like
    transformed y data
    """
    if len(initSpectrumY) > 0 and len(initSpectrumX) > 0:
        Ydata = initSpectrumY
        Xdata = initSpectrumX
    else:
        raise FileNotFoundError
    yf = scipy.fftpack.fft(Ydata)
    xf = np.linspace(Xdata[0], Xdata[-1], len(Xdata))
    return xf, yf


def gaussian_window(t, tau, standardDev, order):
    """
    Returns a simple gaussian window of given parameters evaulated at t.

    Parameters
    ----------
    t: array-like
    input array to perform window on

    tau: float
    center of gaussian window

    standardDev: float
    FWHM of given gaussian

    order: float
    order of gaussian window. If not even it's incremented by 1.

    Returns
    -------
    array : array-like
    nth order gaussian window with params above

    """
    if order % 2 == 1:
        order += 1
    std = standardDev/(2 * (np.log(2)**(1 / order)))
    return np.exp(-((t - tau)**order)/(std**order))


def cut_gaussian(initSpectrumX, initSpectrumY, spike, sigma, win_order):
    """
    Applies gaussian window with the given params.

    Parameters
    ----------
    initSpectrumX: array-like
    x-axis data

    initSpectrumY: array-like
    y-axis data

    spike: float
    center of gaussian window

    sigma: float
    Full width at half max

    Returns
    -------

    Ydata: array-like
    the windowed y values
    """
    Ydata = initSpectrumY * gaussian_window(initSpectrumX, tau=spike, standardDev=sigma, order=win_order)
    return Ydata


def ifft_method(initSpectrumX, initSpectrumY, interpolate=True):
    """
    Perfoms IFFT on data

    Parameters
    ----------

    initSpectrumX: array-like
    the x-axis data

    initSpectrumY: array-like
    the y-axis data

    interpolate: bool
    if True perform a linear interpolation on dataset before transforming

    Returns
    -------
    xf: array-like
    the transformed x data

    yf: array-like
    transformed y data

    """
    if len(initSpectrumY) > 0 and len(initSpectrumX) > 0:
        Ydata = initSpectrumY
        Xdata = initSpectrumX
    else:
        raise ValueError
    N = len(Xdata)
    if interpolate:
        Xdata, Ydata = fourier_interpolate(Xdata, Ydata)
    xf = np.fft.fftfreq(N, d=(Xdata[1]-Xdata[0])/(2*np.pi))
    yf = np.fft.ifft(Ydata)
    return xf, yf


def args_comp(
        initSpectrumX, initSpectrumY, reference_point=0, fitOrder=5, showGraph=False
        ):
    """
    Calculates the phase of complex dataset then unwrap by changing deltas between
    values to 2*pi complement. At the end, fit a polynomial curve to determine
    dispersion coeffs.

    Parameters
    ----------

    initSpectrumX: array-like
    the x-axis data

    initSpectrumY: array-like
    the y-axis data

    reference_point: float
    the reference point to calculate order

    fitOrder: int
    degree of polynomial to fit data [1, 5]

    showGraph: bool
    if True returns a matplotlib plot and pauses execution until closing the window

    Returns
    -------

    dispersion: array-like
    [GD, GDD, TOD, FOD, QOD]

    dispersion_std: array-like
    [GD_std, GDD_std, TOD_std, FOD_std, QOD_std]

    fit_report: lmfit report
    """
    Xdata = initSpectrumX-reference_point
    angles = np.angle(initSpectrumY)
    # shifting to [0, 2pi] if necessary
    # angles = (angles + 2 * np.pi) % (2 * np.pi)
    Ydata = np.unwrap(angles, axis=0)
    if fitOrder == 5:
        fitModel = Model(polynomialFit5)
        params = fitModel.make_params(b0=0, b1=1, b2=1, b3=1, b4=1, b5=1)
        result = fitModel.fit(Ydata, x=Xdata, params=params, method='leastsq')
    elif fitOrder == 4:
        fitModel = Model(polynomialFit4)
        params = fitModel.make_params(b0=0, b1=1, b2=1, b3=1, b4=1)
        result = fitModel.fit(Ydata, x=Xdata, params=params, method='leastsq')
    elif fitOrder == 3:
        fitModel = Model(polynomialFit3)
        params = fitModel.make_params(b0=0, b1=1, b2=1, b3=1)
        result = fitModel.fit(Ydata, x=Xdata, params=params, method='leastsq')
    elif fitOrder == 2:
        fitModel = Model(polynomialFit2)
        params = fitModel.make_params(b0=0, b1=1, b2=1)
        result = fitModel.fit(Ydata, x=Xdata, params=params, method='leastsq')
    elif fitOrder == 1:
        fitModel = Model(polynomialFit1)
        params = fitModel.make_params(b0=0, b1=1)
        result = fitModel.fit(Ydata, x=Xdata, params=params, method='leastsq')
    else:
        raise ValueError('Order is out of range, please select from [1,5]')
    try:
        dispersion, dispersion_std = lmfit_disp(result.params.items())
        dispersion = dispersion[1:]
        dispersion_std = dispersion_std[1:]
        for idx in range(len(dispersion)):
            dispersion[idx] = dispersion[idx] * factorial(idx+1)
            dispersion_std[idx] = dispersion_std[idx] * factorial(idx+1)
        while len(dispersion) < 5:
            dispersion.append(0)
            dispersion_std.append(0)
        fit_report = result.fit_report()
        if showGraph:
            fig = plt.figure(figsize=(7, 7))
            fig.canvas.set_window_title('Phase')
            plt.plot(Xdata, Ydata, 'o', label='dataset')
            plt.plot(Xdata, result.best_fit, 'r--', label='fitted')
            plt.legend()
            plt.grid()
            plt.show()
        return dispersion, dispersion_std, fit_report
    except Exception as e:
        return [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], e



import matplotlib
from matplotlib import pyplot as plt
import numpy as np
from pysprint.utils import calc_envelope
from pysprint.core.peak import SelectButton


class DraggableEnvelope:

    epsilon = 5 # max absolute pixel distance to count as a hit

    def __init__(self, x, y, mode='l'):
        matplotlib.rcParams["toolbar"] = "toolmanager"
        self.fig, self.ax = plt.subplots()
        self.x = x
        self.y = y
        self.mode = mode
        if self.mode == 'l':
            self.envelope, self.y_env, self.loc = calc_envelope(
                self.y, np.arange(len(self.y)), 'l'
                )
            plt.title('Adjust the lower envelope.')
        elif self.mode == 'u':
            self.envelope, self.y_env, self.loc = calc_envelope(
                self.y, np.arange(len(self.y)), 'u'
                )
            plt.title('Adjust the upper envelope.')
        else:
            raise ValueError('mode must be u or l.')
        self._ind = None # the active point index
        self.x_env = self.x[self.loc]

        self.basedata, = self.ax.plot(self.x, self.y)
        self.lines, = self.ax.plot(self.x, self.envelope, 'r')
        self.peakplot, = self.ax.plot(self.x_env, self.y_env, 'ko')
        self.fig.canvas.mpl_connect('button_press_event', self.button_press_callback)
        self.fig.canvas.mpl_connect('key_press_event', self.key_press_callback)
        self.fig.canvas.mpl_connect('draw_event', self.draw_callback)
        self.fig.canvas.mpl_connect('button_release_event', self.button_release_callback)
        self.fig.canvas.mpl_connect('motion_notify_event', self.motion_notify_callback)
        tm = self.fig.canvas.manager.toolmanager
        tm.add_tool('Toggle recording', SelectButton)
        self.fig.canvas.manager.toolbar.add_tool(
            tm.get_tool('Toggle recording'), "toolgroup"
            )
        self.my_select_button = tm.get_tool('Toggle recording')
        plt.grid()
        plt.show()


    def button_release_callback(self, event):
        '''whenever a mouse button is released'''
        if event.button != 1:
            return
        self._ind = None


    def get_ind_under_point(self, event):
        '''Get the index of the selected point within the given epsilon tolerance.'''

        # We use the pixel coordinates, because the axes are usually really
        # differently scaled.
        x, y = self.peakplot.get_data()
        xy_pixels = self.ax.transData.transform(np.vstack([x,y]).T)
        xpix, ypix = xy_pixels.T

        # return the index of the point iff within epsilon distance.
        d = np.hypot(xpix - event.x, ypix - event.y)
        indseq, = np.nonzero(d == d.min())
        ind = indseq[0]

        if d[ind] >= self.epsilon:
            ind = None
        return ind


    def button_press_callback(self, event):
        '''whenever a mouse button is pressed we get the index'''
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        self._ind = self.get_ind_under_point(event)


    def key_press_callback(self, event):
        '''whenever a key is pressed'''
        if not event.inaxes:
            return
        if self.my_select_button.toggled:
            if event.key == 'd':
                ind = self.get_ind_under_point(event)
                if ind is not None:
                    self.x_env = np.delete(self.x_env,
                                             ind)
                    self.y_env = np.delete(self.y_env, ind)
                    self._interpolate()
                    self.peakplot.set_data(self.x_env, self.y_env)
                    self.lines.set_data(self.x, self.envelope)
            elif event.key == 'i':
                self.y_env = np.append(self.y_env, event.ydata)
                self.x_env = np.append(self.x_env, event.xdata)
                self._interpolate()
                self.peakplot.set_data(self.x_env, self.y_env)
                self.lines.set_data(self.x, self.envelope)
            if self.peakplot.stale:
                self.fig.canvas.draw_idle()

    def _interpolate(self):
        '''Upon modifying the datapoints we need to sort values and 
        recalculate the interpolation.'''
        idx = np.argsort(self.x_env)
        self.y_env, self.x_env = self.y_env[idx], self.x_env[idx]
        self.envelope = np.interp(self.x, self.x_env, self.y_env)

    def get_data(self):
        '''
        Only returns the y values accordingly
        '''
        if self.mode == 'l':
            return self.y-self.envelope
        elif self.mode == 'u':
            return self.y/self.envelope

    # need to implement this.. the background is wrong
    def draw_callback(self, event):
        self.background = self.fig.canvas.copy_from_bbox(self.ax.bbox)
        self.ax.draw_artist(self.peakplot)
        self.ax.draw_artist(self.lines)

    def motion_notify_callback(self, event):
        '''on mouse movement we move the selected point'''
        if self._ind is None:
            return
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        if self.my_select_button.toggled:
            x, y = event.xdata, event.ydata
            self.x_env[self._ind], self.y_env[self._ind] = x, y
            self._interpolate()
            self.peakplot.set_data(self.x_env, self.y_env)
            self.lines.set_data(self.x, self.envelope)

            # TODO: make faster drawing with background
            # self.fig.canvas.restore_region(self.background)

            self.ax.draw_artist(self.lines)
            self.ax.draw_artist(self.peakplot)
            self.fig.canvas.blit(self.ax.bbox)
            self.fig.canvas.draw()




"""
Methods for manipulating the loaded data
"""
import numpy as np
from scipy.signal import find_peaks, savgol_filter, gaussian, convolve, find_peaks_cwt
from scipy.interpolate import interp1d
from pysprint.utils import findNearest, _handle_input, between, _maybe_increase_before_cwt



def cwt(x, y, ref, sam, width, floor_thres=0.1):
	Xdata, Ydata = _handle_input(
		x, y, ref, sam
		)
	idx = find_peaks_cwt(Ydata, np.arange(1, width))
	if _maybe_increase_before_cwt(Ydata, tolerance=floor_thres):
		Ydata += 2
	idx2 = find_peaks_cwt(1/Ydata, np.arange(1, width))

	return Xdata[idx], Ydata[idx]-2, Xdata[idx2], Ydata[idx2]-2

def savgol(
	initSpectrumX, initSpectrumY, referenceArmY, sampleArmY, window=101,
	order=3):

	Xdata, Ydata = _handle_input(
		initSpectrumX, initSpectrumY, referenceArmY, sampleArmY
		)
	xint, yint = interpolate_data(Xdata, Ydata, [], [])
	if window > order:
		try:
			if window % 2 == 1:
				fil = savgol_filter(yint, window_length = window, polyorder = order)
				return xint, fil
			else:
				fil = savgol_filter(yint, window_length = window + 1, polyorder = order)
				return xint, fil
		except Exception as e:
			print(e)
	else:
		raise ValueError('Window must be bigger than order currently: {window} and {order})')

def find_peak(
	initSpectrumX, initSpectrumY, referenceArmY, sampleArmY,
	proMax=1, proMin=1, threshold=0.1, except_around=None, 
	rebase=None):   
	if except_around is not None and len(except_around) != 2:
		raise ValueError('Invalid except_around arg. Try [start, stop].')
	if except_around is not None:
		try:
			float(except_around[0])
			float(except_around[1])
		except ValueError:
			raise ValueError(f'Invalid except_around arg. Only numeric values are allowed.')
	if rebase is not None:
		pass # TODO
	Xdata, Ydata = _handle_input(initSpectrumX, initSpectrumY, referenceArmY, sampleArmY)
	maxIndexes, _ = find_peaks(Ydata, prominence=proMax) 
	Ydata_rec = 1/Ydata
	minIndexes, _ = find_peaks(Ydata_rec, prominence=proMin)
	min_idx = []
	max_idx = []
	for idx in maxIndexes:
		if between(Xdata[idx], except_around) or np.abs(Ydata[idx]) > threshold:
			max_idx.append(idx)
	for idx in minIndexes:
		if between(Xdata[idx], except_around) or np.abs(Ydata[idx]) > threshold:
			min_idx.append(idx)

	if len(Xdata[max_idx]) != len(Ydata[max_idx]) or len(Xdata[min_idx]) != len(Ydata[min_idx]):
		raise ValueError('Something went wrong, try to cut the edges of data.')

	return Xdata[max_idx], Ydata[max_idx], Xdata[min_idx], Ydata[min_idx]

def convolution(
	initSpectrumX, initSpectrumY, referenceArmY, sampleArmY,
	win_len ,standev=200):
	Xdata, Ydata = _handle_input(
		initSpectrumX, initSpectrumY, referenceArmY, sampleArmY
		)
	if win_len < 0 or win_len > len(Xdata):
		raise ValueError('Window length must be 0 < window_length < len(x)')

	xint, yint = interpolate_data(Xdata, Ydata, [], [])
	window = gaussian(win_len, std=standev)
	smoothed = convolve(yint, window/window.sum(), mode='same')
	return xint, smoothed

def interpolate_data(initSpectrumX, initSpectrumY, referenceArmY, sampleArmY):
	Xdata, Ydata = _handle_input(
		initSpectrumX, initSpectrumY, referenceArmY, sampleArmY
		)
	xint = np.linspace(Xdata[0], Xdata[-1], len(Xdata))
	intp = interp1d(Xdata,Ydata, kind='linear')
	yint = intp(xint)
	return xint, yint


def cut_data(
	initSpectrumX, initSpectrumY, referenceArmY, sampleArmY,
	startValue=-9999, endValue=9999):

	Xdata, Ydata = _handle_input(
		initSpectrumX, initSpectrumY, referenceArmY, sampleArmY
		)
	if startValue < endValue:
		lowItem, lowIndex = findNearest(Xdata, startValue)
		highItem, highIndex = findNearest(Xdata, endValue)
		mask = np.where((Xdata>=lowItem) & (Xdata<=highItem))
		return Xdata[mask], Ydata[mask]
	else:
		pass



''' Testing the main window base functionality. '''

import sys
import unittest
import numpy as np
from PyQt5.QtWidgets import QApplication
from PyQt5.QtTest import QTest
from PyQt5.QtCore import Qt
from pysprint.logic import GeneratorWindow


class TestGen(unittest.TestCase):
	pass
    # def setUp(self):
    #     self.app = QApplication(sys.argv)
    #     self.gen = GeneratorWindow()
    #     self.gen.showMaximized()

    # def tearDown(self):
    #     self.gen.close()

    # def test_gen(self):
    #     g = self.gen.pushButton_4
    #     self.gen.startLine.setText('4')
    #     self.gen.stopLine.setText('3')
    #     QTest.mouseClick(g, Qt.LeftButton)
    #     assert len(self.gen.xAxisData) == 0
    #     assert len(self.gen.yAxisData) == 0

    # def test_gen1(self):
    #     g = self.gen.pushButton_4
    #     QTest.mouseClick(g, Qt.LeftButton)
    #     assert len(self.gen.xAxisData) != 0
    #     assert len(self.gen.yAxisData) != 0
    #     self.gen.generate_data()


if __name__ == "__main__":
    unittest.main()



''' Testing the main window base functionality. '''

import sys
import unittest

import numpy as np

from PyQt5.QtWidgets import QApplication
from PyQt5.QtTest import QTest
from PyQt5.QtCore import Qt
from pysprint.logic import MainProgram



class Test(unittest.TestCase):
    pass
    # def setUp(self):
    #     self.app = QApplication(sys.argv)
    #     self.my_interface = MainProgram()
    #     self.my_interface.showMaximized()

    # def tearDown(self):
    #     # self.my_interface.close()
    #     pass

    # def test_basic_func(self):
    #     x, y = np.loadtxt('test_peak.txt', unpack=True, delimiter=',')
    #     self.my_interface.a = x
    #     self.my_interface.b = y
    #     self.my_interface.redraw_graph()
    #     self.my_interface.fill_table()
    #     self.my_interface.track_stats()
    #     swap = self.my_interface.swapButton
    #     QTest.mouseClick(swap, Qt.LeftButton)
    #     np.testing.assert_array_equal(x, self.my_interface.b)
    #     np.testing.assert_array_equal(y, self.my_interface.a)
    #     self.my_interface.msg_output('test')
    #     self.my_interface.drop_arms_func()
    #     self.my_interface.show_window()
    #     self.my_interface.gauss_cut_func()
    #     self.my_interface.ifft_handler()
    #     self.my_interface.fft_handler()
    #     # with self.assertRaises(ValueError):
    #     #     self.my_interface.get_it()
    #     # with self.assertRaises(ValueError):
    #     #     self.my_interface.cff_fit_optimizer()

    # def test_basic_errors(self):
    #     with self.assertRaises(ValueError):
    #         self.my_interface.a = np.array([1,3,4])
    #         self.my_interface.b = np.array([])
    #         self.my_interface.redraw_graph()
    #     self.my_interface.reset_all()



if __name__ == "__main__":
    unittest.main()



from abc import abstractmethod, ABCMeta

__all__ = ['DatasetBase']

class DatasetBase(metaclass=ABCMeta):

	@abstractmethod
	def __init__(self):
		pass

	@property
	@abstractmethod
	def data(self):
		pass

	@abstractmethod
	def run(self):
		pass

	@abstractmethod
	def show(self):
		pass


from functools import wraps

import numpy as np
from scipy.interpolate import interp1d
import scipy.stats as st

__all__ = ['scipy_disp', 'lmfit_disp', 'findNearest', 'find_closest',
           '_handle_input', 'print_disp', 'fourier_interpolate',
           'between', 'get_closest', 'run_from_ipython',
           'calc_envelope', 'measurement', '_maybe_increase_before_cwt']


def _maybe_increase_before_cwt(y, tolerance=0.05):
	y = np.asarray(y)
	value = y.min()
	if value - tolerance >= 0:
		return False
	if value < 0 and np.abs(value) - tolerance > 0:
		return True
	return True


def calc_envelope(x, ind, mode='u'):
    '''Source: https://stackoverflow.com/a/39662343/11751294
    '''
    x_abs = np.abs(x)
    if mode == 'u':
    	loc = np.where(np.diff(np.sign(np.diff(x_abs))) < 0)[0] + 1
    elif mode == 'l':
    	loc = np.where(np.diff(np.sign(np.diff(x_abs))) > 0)[0] + 1
    else:
    	raise ValueError('mode must be u or l.')
    peak = x_abs[loc]
    envelope = np.interp(ind, loc, peak)
    return envelope, peak, loc

def run_from_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False

def get_closest(xValue, xArray, yArray):
	idx = (np.abs(xArray - xValue)).argmin()
	value = xArray[idx]
	return value, yArray[idx], idx

def between(val, except_around):
	if except_around is None:
		return False
	elif len(except_around) != 2:
		raise ValueError(f'Invalid interval. Try [start, end] instead of {except_around}')
	else:
		lower = float(min(except_around))
		upper = float(max(except_around))
	if val <= upper and val >= lower:
		return True
	return False

def scipy_disp(r):
	for idx in range(len(r)):
		dispersion[idx] = dispersion[idx] * factorial(idx+1)
		dispersion_std[idx] = dispersion_std[idx] * factorial(idx+1)
	return dispersion, dispersion_std


def lmfit_disp(r):
	dispersion, dispersion_std = [], []
	for name, par in r:
		dispersion.append(par.value)
		dispersion_std.append(par.stderr)
	return dispersion, dispersion_std


def measurement(array, confidence=0.95):
	"""
	Give the measurement results with condifence interval
	assuming the standard deviation is unknown.

	Parameters
	----------

	array : ndarray
		The array containing the measured values

	confidence : float, optional
		The desired confidence level. Must be between 0 and 1.

	Returns
	-------

	mean: float
		The mean of the given array

	conf: tuple-like (interval)
		The confidence interval

	Example(s)
	---------
	>>> import numpy as np
	>>> from pysprint.utils import measurement
	>>> a = np.array([123.783, 121.846, 122.248, 125.139, 122.569])
	>>> mean, interval = measurement(a, 0.99)
	123.117000 ± 2.763022
	>>> mean
	123.117
	>>> interval
	(120.35397798230359, 125.88002201769642)

	Notes
	-----

	I decided to print the results immediately, because people often don't use
	it for further code. Of course, they are also returned if needed.

	"""
	mean = np.mean(array)
	conf = st.t.interval(confidence, len(array)-1, loc=mean, scale=st.sem(array))
	print(f'{mean:5f} ± {(mean-conf[0]):5f}')
	return mean, conf


def findNearest(array, value):
	#Finds the nearest element to the given value in the array
	#returns tuple: (element, element's index)
	
    array = np.asarray(array)
    idx = (np.abs(value - array)).argmin()
    return array[idx], idx


def _handle_input(initSpectrumX, initSpectrumY, referenceArmY, sampleArmY):
	"""
	Instead of handling the inputs in every function, there is this private method.

	Parameters
	----------

	initSpectrumX: array-like
	x-axis data

	initSpectrumY: array-like
	y-axis data

	referenceArmY, sampleArmY: array-like
	reference and sample arm spectrum evaluated at initSpectrumX

	Returns
	-------
	initSpectrumX: array-like
	unchanged x data

	Ydata: array-like
	the transformed y data

	"""
	if (len(initSpectrumX) > 0) and (len(referenceArmY) > 0) and (len(sampleArmY) > 0):
		Ydata = (initSpectrumY - referenceArmY - sampleArmY) / (2 * np.sqrt(referenceArmY * sampleArmY))
	elif (len(referenceArmY) == 0) or (len(sampleArmY) == 0):
		Ydata = initSpectrumY
	elif len(initSpectrumX) == 0:
		raise ValueError('Please load the spectrum!\n')
	elif len(initSpectrumY) == 0:
		raise ValueError('Please load the spectrum!\n')
	else:
		raise TypeError('Input types are wrong.\n')
	return initSpectrumX,  Ydata


def find_closest(xValue, xArray, yArray):
	idx = (np.abs(xArray - xValue)).argmin()
	value = xArray[idx]
	return value, yArray[idx]


def print_disp(f):
    @wraps(f)
    def wrapping(*args, **kwargs):
        disp, disp_std, stri = f(*args, **kwargs)
        labels = ('GD', 'GDD','TOD', 'FOD', 'QOD')
        for i, (label, disp_item, disp_std_item) in enumerate(zip(labels, disp, disp_std)):
             print(f'{label} = {disp_item:.5f} ± {disp_std_item:.5f} fs^{i+1}')
        return disp, disp_std, stri
    return wrapping


def fourier_interpolate(x, y):
    ''' Simple linear interpolation for FFTs'''
    xs = np.linspace(x[0], x[-1], len(x))
    intp = interp1d(x, y, kind='linear', fill_value='extrapolate')
    ys = intp(xs)
    return xs, ys



"""
- currently compatibility with GUI is dropped.. we should fix that.
"""
from math import factorial

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from tqdm import tqdm

from pysprint.utils import findNearest as find_nearest
from pysprint.core.evaluate import cos_fit1, cos_fit2, cos_fit3, cos_fit4, cos_fit5

_func_config = {
	1 : cos_fit1,
	2 : cos_fit2,
	3 : cos_fit3,
	4 : cos_fit4,
	5 : cos_fit5
	}

class FitOptimizer:
	"""Class to help achieve better fitting results."""
	def __init__(self, x, y, ref, sam, reference_point, max_order=3):
		self.x = x
		self.y = y
		self.ref = ref
		self.sam = sam
		if not isinstance(self.x, np.ndarray):
			self.x = np.asarray(self.x)
		if not isinstance(self.y, np.ndarray):
			self.y = np.asarray(self.y)
		if not isinstance(self.ref, np.ndarray):
			try:
				self.ref = np.asarray(self.ref)
			except (ValueError, TypeError):
				pass # we ignore because it might be optional if y is already normalized
		if not isinstance(self.sam, np.ndarray):
			try:
				self.sam = np.asarray(self.sam)
			except (ValueError, TypeError):
				pass # we ignore because it might be optional if y is already normalized
		if len(self.ref) == 0:
			self._y_norm = self.y
		else:
			self._y_norm = (self.y - self.ref - self.sam)/(2*np.sqrt(self.sam*self.ref))
		self.reference_point = reference_point
		self.x -= self.reference_point
		self.func = cos_fit1
		self.p0 = [1, 1, 1, 1]
		self.popt = self.p0
		self._init_set = False
		self.counter = 0
		self.curr_order = 1
		self.max_order = max_order
		self.rest = None
		self.figure = plt.figure

	def __del__(self):
		self.reference_point = 0
		self.x = None
		self.y = None
		self.ref = None
		self.sam = None

	def set_final_guess(self, GD, GDD=None, TOD=None, FOD=None, QOD=None):
		self.p0[3] = GD
		self.rest = []
		if GDD is not None:
			self.rest.append(GDD/2)
		if TOD is not None:
			self.rest.append(TOD/6)
		if FOD is not None:
			self.rest.append(FOD/24)
		if QOD is not None:
			self.rest.append(QOD/120)

	@property
	def user_guess(self):
		return self.rest

	# TODO: Make this update the plot, not redraw
	def update_plot(self):
		plt.clf()
		plt.plot(self.x, self._y_norm)
		plt.plot(self._x_curr, self._y_curr, 'k', label='Affected data')
		plt.plot(self._x_curr, self.func(self._x_curr, *self.popt), 'r--', label='Fit')
		plt.grid()
		plt.legend(loc='upper left')
		plt.draw()
		plt.show()

	def set_initial_region(self, percent):
		""" Determines the initial region to fit"""
		self._init_set = True
		_, idx = find_nearest(self.x, 0)
		self._upper_bound = np.floor(idx + (percent/2)*(len(self.x) + 1))
		self._lower_bound = np.floor(idx - (percent/2)*(len(self.x) + 1))
		self._upper_bound = self._upper_bound.astype(int)
		self._lower_bound = self._lower_bound.astype(int)
		if self._lower_bound < 0:
			self._lower_bound = 0
		if self._upper_bound > len(self.x):
			self._upper_bound = len(self.x)
		self._x_curr = self.x[self._lower_bound:self._upper_bound]
		self._y_curr = self._y_norm[self._lower_bound:self._upper_bound]

	def _step_up_func(self):
		"""
		Change the function to fit. Starts with first order
		and stepping up until max_order.
		"""
		if self.curr_order == self.max_order:
			return
		try:
			self.func = _func_config[self.curr_order + 1]
		except KeyError as e:
			e.args = (e.args[0], 'Order must be in [1, 5].')
			raise
		try:
			self.p0 = np.append(self.p0, self.rest[self.curr_order-1])
		except (IndexError, ValueError):
			self.p0 = np.append(self.p0, 1)
		self.curr_order += 1

	def _extend_region(self, extend_by=0.1):
		""" Extends region of fit"""
		self._new_lower = np.floor(self._lower_bound - extend_by*len(self.x))
		self._new_upper = np.floor(self._upper_bound + extend_by*len(self.x))
		self._new_lower = self._new_lower.astype(int)
		self._new_upper = self._new_upper.astype(int)
		self._lower_bound = self._new_lower
		self._upper_bound = self._new_upper
		if self._new_lower < 0:
			self._new_lower = 0
		if self._new_upper > len(self.x):
			self._new_upper = len(self.x)
		self._x_curr = self.x[self._new_lower:self._new_upper]
		self._y_curr = self._y_norm[self._new_lower:self._new_upper]

	def _finetune(self):
		"""
		Changes the last parameter randomly, we might get lucky..
		"""
		self.p0[-1] = float(np.random.uniform(-1, 1, 1)) * self.p0[-1]

	def _make_fit(self):
		try:
			if len(self._x_curr) == len(self.x):
				return True
			self.popt, self.pcov = curve_fit(
				self.func, self._x_curr, self._y_curr, p0=self.p0
				)
			self.p0 = self.popt
		except RuntimeError:
			self._finetune()
			self.popt, self.pcov = curve_fit(
				self.func, self._x_curr, self._y_curr, p0=self.p0
				)


	def _fit_goodness(self):
		"""
		Coeffitient of determination a.k.a. r^2
		"""
		residuals = self._y_curr - self.func(self._x_curr, *self.popt)
		ss_res = np.sum(residuals**2)
		ss_tot = np.sum((self._y_curr - np.mean(self._y_curr))**2)
		return 1 - (ss_res / ss_tot)

	def result_wrapper(self):
		labels = ('GD', 'GDD', 'TOD', 'FOD', 'QOD')
		params = self.p0[3:]
		for i, (label, param) in enumerate(zip(labels, params)):
			print(f'{label} = {(params[i]*factorial(i+1)):.5f} fs^{i + 1}')

	def run_loop(
		self, r_extend_by, r_threshold,
		max_tries=5000, show_endpoint=True
		):
		pbar = tqdm(total=max_tries)
		if not self._init_set:
			raise ValueError('Set the initial conditions.')
		self._make_fit()
		while self._fit_goodness() > r_threshold:
			self._extend_region(r_extend_by)
			self._make_fit()
			self.counter += 1
			self._step_up_func()
			pbar.update(1)
			if self._make_fit() is True:
				pbar.close()
				if show_endpoint:
					self.update_plot()
				self.result_wrapper()
				print(f'with r^2 = {(self._fit_goodness()):.5f}.')
				return self.popt
			if self.counter == max_tries:
				if show_endpoint:
					self.update_plot()
				pbar.close()
				print(f'''Max tries ({max_tries}) reached.. try another initial params''')
				return np.zeros_like(self.popt)

		while self._fit_goodness() < r_threshold:
			self._make_fit()
			# self._finetune()
			self.counter += 1
			pbar.update(1)
			if self.counter == max_tries:
				if show_endpoint:
					self.update_plot()
				pbar.close()
				print(f'''\nMax tries ({max_tries}) reached.. try another initial params''')
				return np.zeros_like(self.popt)



import pandas as pd

__all__ = ['convert_df']

def convert_df(
	filename, ifg, ref=None, sam=None, delimiter=';', decimal=',',
	skiprows=0, index=False, invert_axis=False, fmt=None
	):
	"""
	Convert and save given interferogram datafile (either single 
	or along with arms) to PySprint GUI - loadable format.

	Parameters
	----------

	filename: string
	Name of the generated output file.

	ifg: string
	Path to the original interferogram file.

	ref: string, optional
	Path to the reference arm file.

	sam: string, optional
	Path to the sample arm file.

	delimiter: string, optional
	The delimiter in the original interferogram files.
	Default is ';'.

	decimal: string, optional
    Character recognized as decimal separator in the original dataset. 
    Often ',' for European data.
	Default is ','.

	skiprows: int, optional
	Number of rows to skip at the beginning of the file.
	Default is 0.

	index: bool, optional
	Write row numbers (index). Keep this False to make sure it's safely loadable.
	Default is False.

	invert_axis: bool, optional
	Whether to treat first columns as y values in the original interferogram files.
	Default is False.

	fmt: string, optional
	Format string for floating point numbers.
	More info at https://docs.python.org/3.4/library/string.html#format-string-syntax
	
	Returns
	-------
	None
	"""
	ifg_data = pd.read_csv(
		ifg, sep=delimiter, decimal=decimal, skiprows=skiprows, names=['x', 'y']
		)
	if (sam is not None and ref is not None):
		sam_data = pd.read_csv(
			sam, sep=delimiter, decimal=decimal, skiprows=skiprows, names=['x', 'y']
			)
		samy = sam_data['y'].values if not invert_axis else sam_data['x'].values
		ref_data = pd.read_csv(
			ref, sep=delimiter, decimal=decimal, skiprows=skiprows, names=['x', 'y']
			)
		refy = ref_data['y'].values if not invert_axis else ref_data['x'].values
	x = ifg_data['x'].values
	y = ifg_data['y'].values
	if invert_axis:
		x, y = y, x
	try:
		fulldf = pd.DataFrame({'x': x, 'y': y, 'ref': refy, 'sam': samy})
	except NameError:
		fulldf = pd.DataFrame({'x': x, 'y': y})
	fulldf.to_csv(filename, sep=',', decimal='.', index=index, float_format=fmt)
	return
